// QDN.cpp: implementation of the QRDN and QDN classes.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Query32.h"
#include "QDN.h"
#include "QDSManager.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// QRDNAttr - An RDN attribute (an attribute with a single value)
//////////////////////////////////////////////////////////////////////

QRDNAttr::QRDNAttr( QAttrT *attrt_arg, char *ldap_val_ptr, unsigned long len )
{
  attrt = attrt_arg;
  rdn_val = new QValue( attrt->Syntax(), ldap_val_ptr, len );
}

QRDNAttr::~QRDNAttr()
{
  delete rdn_val;
}

void QRDNAttr::Print( CString &str )
{
  WIN32_GUFF();

  CString val_str;
  rdn_val->Print( val_str );
  str = attrt->Name() + "=" + val_str;
}

BOOL QRDNAttr::operator==( QRDNAttr &ra )
{
  if ( this->Type() == ra.Type() && *this->Value() == *ra.Value() )
    return TRUE;

  return FALSE;
}

//////////////////////////////////////////////////////////////////////
// QRDN - Relative Distinguished Name
//////////////////////////////////////////////////////////////////////

QRDN::QRDN()
{
  WIN32_GUFF();
}

QRDN::QRDN( Oidtable *table, const CString &rdn_str )
{
  WIN32_GUFF();

  CString str = rdn_str;
  CStringList str_list;

	//
  // Split the RDN string into RDN value components
  BOOL more = FALSE;
	str.TrimLeft();

	while ( str.IsEmpty() == FALSE )
		{
			int pos = str.Find( '+' );
			if ( pos > 0 )
				{
					if ( str[ pos - 1 ] == '\\' )
						// Ignore escaped '+' characters
						continue;

					CString a = str.Left( pos );
					a.TrimLeft();
					a.TrimRight();
					str_list.AddTail( a );

					// Remove read RDN 
					str = str.Right( str.GetLength() - pos - 1 );
				}
			else
				{
					CString a = str;
					a.TrimLeft();
					a.TrimRight();
					str_list.AddTail( a );
					str.Empty();
				}
		}

	//
  // Got a sequence of the RDN attribute components, now parse each of them
  POSITION pos = str_list.GetHeadPosition();
  while ( pos != NULL )
    {
      CString &comp_str = str_list.GetNext( pos );
      int pos = comp_str.Find( '=' );
      if ( pos > 0 )
        {
          // Get the attribute and value strings
          CString attr_str = comp_str.Left( pos );
          attr_str.TrimRight();

          CString val_str = comp_str.Right( comp_str.GetLength() - pos - 1 );
          val_str.TrimLeft();

          QAttrT *attrt = table->str2at( attr_str );
          if ( attrt != NULL )
            {
              char *ldap_val_str = strdup( (LPCSTR) val_str );
              QRDNAttr *attr = new QRDNAttr( attrt, ldap_val_str, strlen( ldap_val_str ) + 1 );
              rdn_value_list.AddTail( attr );
              free( (void *) ldap_val_str );
            }
#ifdef _DEBUG
          else
            {
              CString msg;
              msg.Format( "Failed to recognize attribute type: %s\n", attr_str );
              OutputDebugString( msg );
            }
#endif              
        }
    }
}

QRDN::~QRDN()
{
  WIN32_GUFF();

  POSITION pos = rdn_value_list.GetHeadPosition();
  while ( pos != NULL )
    {
      QRDNAttr *comp = rdn_value_list.GetNext( pos );
      delete comp;
    }
}

BOOL QRDN::GetRDNString( CString &str )
{
  WIN32_GUFF();

  str.Empty();
  POSITION pos = rdn_value_list.GetHeadPosition();
  while ( pos != NULL )
    {
      CString attr_str;
      QRDNAttr *attr = rdn_value_list.GetNext( pos );
      attr->Print( attr_str );

      if ( str.IsEmpty() )
        str = attr_str;
      else
        str = attr_str + " + " + str;
    }

  return str.IsEmpty();
}

BOOL QRDN::GetFriendlyRDNString( CString &str )
{
  WIN32_GUFF();

  str.Empty();
  POSITION pos = rdn_value_list.GetHeadPosition();
  while ( pos != NULL )
    {
      CString val_str;
      QRDNAttr *attr = rdn_value_list.GetNext( pos );
			QValue *value = attr->Value();

      value->Print( val_str );
      if ( str.IsEmpty() )
        str = val_str;
      else
        str = val_str + ", " + str;
    }

  return str.IsEmpty();
}

BOOL QRDN::operator==( QRDN &r )
{
  WIN32_GUFF();

	if ( rdn_value_list.GetCount() != r.rdn_value_list.GetCount() )
		return FALSE;

  PtrList<QRDNAttr> r_copy;
  POSITION p2 = r.rdn_value_list.GetHeadPosition();
  while ( p2 != NULL )
    r_copy.AddTail( r.rdn_value_list.GetNext( p2 ) );

  POSITION p1 = this->rdn_value_list.GetHeadPosition();
  while ( p1 != NULL )
    {
      BOOL got_it = FALSE;
      QRDNAttr *a1 = this->rdn_value_list.GetNext( p1 );
      
      p2 = r_copy.GetHeadPosition();
      while ( p2 != NULL )
        {
          POSITION last_p2 = p2;
          QRDNAttr *a2 = r_copy.GetNext( p2 );
          if ( *a1 == *a2  )
            {
              r_copy.RemoveAt( last_p2 );
              got_it = TRUE;
              break;
            }
        }

      if ( got_it == FALSE)
        return FALSE;
    }

  return TRUE;
}

QRDN &QRDN::operator=( QRDN &rRdn )
{
	POSITION pos = rRdn.GetHeadRDNAttr();
	while ( pos != NULL )
		{
			QRDNAttr *pRDNAttr = rRdn.GetNextRDNAttr( pos );
			QValue *pVal = pRDNAttr->Value() ;

			CString str;
			pVal->Print( str );
			char *pStr = strdup( str );

			QRDNAttr *pCopyRDNAttr =
				new QRDNAttr( pRDNAttr->Type(), pStr, strlen( pStr ) + 1 );

			rdn_value_list.AddTail( pCopyRDNAttr );
			free( pStr );
		}

	return *this;
}

//////////////////////////////////////////////////////////////////////
// QDN - Distinguished Name
//////////////////////////////////////////////////////////////////////
QDN::QDN()
{
	is_set = FALSE;
	conn_id = UNDEF_CONN_ID ;
	ds = NULL;
} 
 
QDN::QDN( QDSManager *ds, long conn_id, CString &dn_str )
{
  WIN32_GUFF();

	is_set = TRUE;
	ds->BindManager()->AddRef( conn_id );
  SetDN( ds, conn_id, dn_str );
}

QDN::~QDN()
{
  WIN32_GUFF();
  UnsetDN();
}

void QDN::UnsetDN()
{
  WIN32_GUFF();

  POSITION pos = rdn_list.GetHeadPosition();
  while ( pos != NULL )
    {
      QRDN *rdn = rdn_list.GetNext( pos );
      delete rdn;
    }

	if ( ds != NULL )
		ds->BindManager()->LoseRef( conn_id );

	conn_id = UNDEF_CONN_ID;
	ds = NULL;
	is_set = FALSE;
}

BOOL QDN::SetDN( QDSManager *ds_arg, long conn_id_arg, CString &str )
{
  WIN32_GUFF();

	UnsetDN();

	if ( ds_arg == NULL || conn_id_arg < 0 )
		// The NULL DN
		return TRUE;

  ds = ds_arg;
	conn_id = conn_id_arg;
	ds->BindManager()->AddRef( conn_id_arg );

  QStringList rdn_str_list;
  CString s = str;

	if ( str.IsEmpty() || str == "" )
		{
			str = "";
			return TRUE;
		}

	//
	// Clean up the input string
	s.TrimLeft();
	s.TrimLeft('/');
	s.TrimRight();
	s.TrimRight('/');

	//
	// Break the DN down into a list of RDN strings
  while ( TRUE )
    {
			//
      // Find end of leading RDN, avoiding escaped
			// separators (the ',' char) along the way
      int i;
      BOOL got_it = FALSE;

      for ( i = 0; i < s.GetLength(); i++ )
        if ( s[ i ] == ',' && ( i == 0 || s[ i - 1 ] != '\\' ) )
          {
            got_it = TRUE;
            break;
          }

      if ( got_it == TRUE )
        {
          if ( i == 0 )
            {
              rdn_str_list.AddTail( "" );
              s = s.Right( s.GetLength() - 1 );
            }
          else if ( i > 0 )
            {
              rdn_str_list.AddTail( s.Left( i ) );
              s = s.Right( s.GetLength() - i - 1 );
            }
        }
      else
        {
          rdn_str_list.AddTail( s );
          break;
        }
    }

	//
  // Create parsed RDN list from list of RDN strings
  POSITION pos = rdn_str_list.GetHeadPosition();
  while ( pos != NULL )
    {
      CString rdn_str = rdn_str_list.GetNext( pos );
      QRDN *rdn = new QRDN( Table(), rdn_str );
      rdn_list.AddTail( rdn );
    }

	is_set = TRUE;
  return TRUE;
}

BOOL QDN::GetDN( CString &str )
{ 
  WIN32_GUFF();

	str = "";
  UINT rdn_count = 0;

  POSITION pos = rdn_list.GetHeadPosition();
	if ( pos == NULL )
		// The NULL DN
		return TRUE;

  while ( pos != NULL )
    {
      QRDN *rdn = rdn_list.GetNext( pos );
      rdn_count++;

      CString rdn_str;
      rdn->GetRDNString( rdn_str );

      if ( rdn_count == 1 )
        str = rdn_str;
      else
        str = str + ", " + rdn_str;
    }

  return ( rdn_count > 0 ? TRUE : FALSE );
}

BOOL QDN::GetFriendlyDN( CString &str )
{
	WIN32_GUFF();

	str = "";
	POSITION pos = rdn_list.GetHeadPosition();
	if ( pos == NULL )
		// The NULL DN
		return TRUE;

	UINT rdn_count = 0;
  while ( pos != NULL )
    {
      QRDN *rdn = rdn_list.GetNext( pos );
      rdn_count++;

      CString rdn_str;
      rdn->GetFriendlyRDNString( rdn_str );

      if ( rdn_count == 1 )
        str = rdn_str;
      else
        str = str + ", " + rdn_str;
    }

  return TRUE;
}

BOOL QDN::GetRDN( QRDN &get_rdn )
{
  WIN32_GUFF();

	if ( GetRDNCount() == 0 )
		return FALSE;

	get_rdn = *rdn_list.GetHead();
  return TRUE;
}

POSITION QDN::GetHeadRDNPosition()
{
	return rdn_list.GetHeadPosition();
}

QRDN *QDN::GetNextRDN( POSITION &rPos )
{
	return rdn_list.GetNext( rPos );
}

QDSManager *QDN::DS()
{
	return ds;
}

long QDN::GetConnID()
{
	return conn_id;
}

Oidtable *QDN::Table()
{
	return ds->Table();
}

BOOL QDN::operator==( QDN &dn2 )
{
  WIN32_GUFF();

  if ( this->GetRDNCount() != dn2.GetRDNCount() ) 
    return FALSE;

  POSITION p1 = this->rdn_list.GetHeadPosition();
  POSITION p2 = dn2.rdn_list.GetHeadPosition();
  while ( p1 != NULL && p2 != NULL )
    {
      QRDN *rdn1 = this->rdn_list.GetNext( p1 );
      QRDN *rdn2 = dn2.rdn_list.GetNext( p2 );

      if ( !( *rdn1 == *rdn2 ) )
        return FALSE;
    }

  return ( p1 == NULL && p2 == NULL ? TRUE : FALSE );
}

QDN &QDN::operator=( QDN &dn_rhs )
{
  WIN32_GUFF();

  CString dn_str;
  dn_rhs.GetDN( dn_str );
  SetDN( dn_rhs.DS(), dn_rhs.GetConnID(), dn_str );
  return *this;
}

